#version 330 compatibility

#info See blog post: <https://mathr.co.uk/blog/2018-11-18_newtons_method_for_periodic_cycles.html>

#include "Progressive2D.frag"
#include "Complex.frag"

// The simplest way to draw something in Fragmentarium,
// is to include the "2D.frag" header.
//
// Now, we can implement a simple 'color' function,
// which for each point in the plane returns a RGB color.
//
// Notice, that you can zoom using the mouse or the keyboard.
// (A,S,D,W,Q,E,1,3)
#define PI 3.141592653
// Hue in radians
vec3 HSVtoRGB(vec3 hsv) {
	/// Implementation based on: http://en.wikipedia.org/wiki/HSV_color_space
	hsv.x = mod(hsv.x,2.*PI);
	int Hi = int(mod(hsv.x / (2.*PI/6.), 6.));
	float f = (hsv.x / (2.*PI/6.)) -float( Hi);
	float p = hsv.z*(1.-hsv.y);
	float q = hsv.z*(1.-f*hsv.y);
	float t = hsv.z*(1.-(1.-f)*hsv.y);
	if (Hi == 0) { return vec3(hsv.z,t,p); }
	if (Hi == 1) { return vec3(q,hsv.z,p); }
	if (Hi == 2) { return vec3(p,hsv.z,t); }
	if (Hi == 3) { return vec3(p,q,hsv.z); }
	if (Hi == 4) { return vec3(t,p,hsv.z); }
	if (Hi == 5) { return vec3(hsv.z,p,q); }
	return vec3(0.);
}

vec4 Mandelbrot_attractor_naive(vec2 c, vec2 z, int period)
{
	vec2 z0 = z;
	vec2 dw;
	for (int j = 0; j < 16; ++j)
	{
		vec2 w = z0;
		dw = vec2(1.0, 0.0);
		for (int l = 0; l < period; ++l)
		{
			dw = 2.0 * cMul(w, dw);
			w = cSqr(w) + c;
		}
		z0 -= cDiv(w - z0, dw - vec2(1.0, 0.0));
	}
	return vec4(z0, dw);
}

vec4 Mandelbrot_attractor(vec2 c, vec2 z, int period)
{
	vec2 z0 = z;
	for (int l = 0; l < period; ++l)
	{
		z0 = cSqr(z0) + c;
	}
	vec2 dw;
	for (int j = 0; j < 16; ++j)
	{
		vec4 G = cConst(1.0);
		vec2 w = z0;
		dw = vec2(1.0, 0.0);
		for (int l = 1; l <= period; ++l)
		{
			dw = 2.0 * cMul(w, dw);
			w = cSqr(w) + c;
			if (period % l == 0 && l < period)
			{
				G = cDiv(G, vec4(w - z0, dw - vec2(1.0, 0.0)));
			}
		}
		G = cMul(G, vec4(w - z0, dw - vec2(1.0, 0.0)));
		z0 -= cDiv(G.xy, G.zw);
	}
	return vec4(z0, dw);
}

vec4 Mandelbrot_attractor_ray(vec2 nucleus, vec2 target, int period)
{
	const int steps = 64;
	vec4 P = vec4(0.0);
	for (int ray = 1; ray <= steps; ++ray)
	{
		P = Mandelbrot_attractor(mix(nucleus, target, float(ray) / float(steps)), P.xy, period);
	}
	return P;
}

vec2 Mandelbrot_nucleus(vec2 c0, int period)
{
	vec2 c = c0;
	for (int j = 0; j < 8; ++j)
	{
		vec4 G = cConst(1.0);
		vec2 w = vec2(0.0);
		vec2 dw = vec2(0.0);
		for (int l = 1; l <= period; ++l)
		{
			dw = 2.0 * cMul(w, dw) + vec2(1.0, 0.0);
			w = cSqr(w) + c;
			
			if (l < period && period % l == 0)
			G = cDiv(G, vec4(w, dw));
		}
		G = cMul(G, vec4(w, dw));
		c -= cDiv(G.xy, G.zw);
	}
	return c;
}

vec2 Mandelbrot_from_cardioid(vec2 dz)
{
	return vec2(1.0, 0.0) - cSqrt(vec2(1.0, 0.0) - 4.0 * dz);
}

vec2 Mandelbrot_to_cardioid(vec2 dz)
{
	return 0.25 * (vec2(1.0, 0.0) - cSqr(vec2(1.0, 0.0) + dz));
}

#group Mandelbrot
uniform int Method; slider[1,1,3]
uniform int Period; slider[1,4,24]
int period = Period;

vec2 formula(vec2 c)
{
	if (Method == 1)
	return Mandelbrot_attractor_naive(c, vec2(0.0), period).zw;
	if (Method == 2)
	return Mandelbrot_attractor      (c, vec2(0.0), period).zw;
	if (Method == 3)
	return Mandelbrot_attractor_ray(Mandelbrot_nucleus(c, period), c, period).zw;
	return c;
}

vec3 color(vec2 z) {
	z = formula(z);
	z = cLog(z);
	float sat = z.x <= 0.0 ? 1.0 : 1.0 / ceil(z.x + 1.0);
	float hue = z.y;
	// grid
	z *= vec2(2.0, 6.0 / PI);
	z -= round(z);
	float val = abs(z.x) < 0.05 ? 0.0 : 1.0;
	sat = abs(z.y) < 0.05 ? 0.0 : sat;
	vec3 v = HSVtoRGB(vec3(hue, sat, val));
	return v;
}

#preset Default
Center = -0.75,0
Zoom = 0.7
Method = 1
Period = 1
#endpreset
